#! /usr/bin/env python3
# -*- coding:utf-8 -*-

# Author   : phuntsok
# Github   : https://github.com/puntsokCN
# Date     : 2019-12-27
# Name     : Word
# Software : VS code
# Note     : 藏文单词元素分析类
# 导入模块  : 无


class Word :
   
    "单词分析类"
    
    def __init__(self, s):
        if len(s) == 0 :
            print("输入字符不可为空！")
        if s[-1] == "་" or s[-1] == "།" :
            if len(s) <= 8 :
                self.__str = s[:-1]
            else :
                print("输入字符过长")
        else :
            if len(s) <= 7 :
                self.__str = s
            else :
                print("输入字符过长")
    
    #############################################################################
    #    私有数据
    __words = [ 'ཀ', 'ཁ', 'ག', 'ང', 'ཅ', 'ཆ', 'ཇ', 'ཉ', 'ཏ', 'ཐ', 'ད', 'ན', 
                'པ', 'ཕ', 'བ', 'མ', 'ཙ', 'ཚ', 'ཛ', 'ཝ', 'ཞ', 'ཟ', 'འ', 'ཡ',
                'ར', 'ལ', 'ཤ', 'ས', 'ཧ', 'ཨ']
    __front = ["ག", "ད", "བ", "མ", "འ"]
    __top = ["ར", "ལ", "ས", "ག", "ད", "ཛ", "བ", "ཧ"]
    __below = ["ྱ", "ྲ", "ླ", "ྭ"]
    __behind = ["ག", "ང", "ད", "ན", "བ", "མ", "འ", "ར", "ལ", "ས"]
    __repeat_behind = ["ད", "ས"]   # ད 虽然也是重后加字，但是现代藏文里已弃用
    __symbol = ['ི',  'ུ',  'ེ' , 'ོ',  'ཽ',  'ྀ',  'ཻ']
    __sub_word_dict = { 'ཀ': 'ྐ',  'ཁ': 'ྑ',  'ག': 'ྒ',  'ང': 'ྔ',  'ཅ': 'ྕ',  'ཆ': 'ྖ',  'ཇ': 'ྗ',  'ཉ': 'ྙ', 
                        'ཏ': 'ྟ',  'ཐ': 'ྠ',  'ད': 'ྡ',  'ན': 'ྣ',  'པ': 'ྤ',  'ཕ': 'ྥ',  'བ': 'ྦ',  'མ': 'ྨ', 
                        'ཙ': 'ྩ',  'ཚ': 'ྪ',  'ཛ': 'ྫ',  'ཝ': 'ྭ',  'ཞ': 'ྮ',  'ཟ': 'ྯ',  'འ': 'ྰ',  'ཡ': 'ྱ',
                        'ར': 'ྲ', 'ལ': 'ླ',  'ཤ': 'ྴ',  'ས': 'ྶ',  'ཧ': 'ྷ',  'ཨ': 'ྸ'}
    __new_dict = dict(zip(__sub_word_dict.values(), __sub_word_dict.keys()))
    __sub_word =['ྐ',   'ྑ',   'ྒ',   'ྔ',   'ྕ',   'ྖ',   'ྗ',   'ྙ',  'ྟ',   'ྠ',   'ྡ',   'ྣ',   'ྤ',   'ྥ',  
                 'ྦ',   'ྨ',   'ྩ',   'ྪ',   'ྫ',   'ྮ',   'ྯ',   'ྰ',   'ྴ',   'ྶ',   'ྷ',   'ྸ']
    
    __special_words = ['ཛྙ', 'ཧྥ', 'གྷ',  'དྷ',  'ཛྷ', 'བྷ']
    
    # 上加字 正字规范
    __top_ra = ['ཀ', 'ག', 'ང', 'ཇ', 'ཉ', 'ཏ', 'ད', 'ན', 'བ', 'མ', 'ཙ', 'ཛ']
    __top_la = ['ཀ', 'ག', 'ང', 'ཇ', 'ཅ', 'ཏ', 'ད', 'པ', 'བ', 'ཧ' ]    
    __top_sa = ['ཀ', 'ག', 'ང', 'ཉ', 'ཏ', 'ད', 'ན', 'པ', 'བ', 'མ', 'ཙ']

    # 下加字 正字规范
    __under_ya = ['ཀ', 'ཁ', 'ག', 'པ', 'ཕ' ,'པ' ,'མ']
    __under_ra = ['ཀ', 'ཁ', 'ག', 'ཏ', 'ད', 'པ', 'ཕ', 'བ', 'མ', 'ཤ', 'ས', 'ཧ']
    __under_la = ['ཀ', 'ག', 'བ', 'ཟ', 'ར', 'ས']
    __under_wa = ['ཀ', 'ཁ', 'ག', 'ཉ', 'ད', 'ཙ', 'ཚ', 'ཞ', 'ཟ', 'ར' ,'ལ', 'ཤ', 'ས', 'ཧ']
    
    # 元素映射:[ 前    上    基    下    音标   后    重后]
    __result = [None, None, None, None, None, None, None]

    
    #############################################################################
    # 核心功能函数：修改 __result 元素  此区域函数能识别现代藏文词的构成元素。无法识别复杂的生僻词、大部分缩写词汇
    
    ## 查找字符串的所有，填充信息至 __result的行为函数，供外调函数 使用
    def __analysis(self):
        "按序分析字符串，并保存信息至 __result"
        all = [self.__if_single(), self.__if_top(), self.__if_below(), self.__if_symbol(), self.__if_pure_word()]
        for func in all :            # 只要字符满足其中之一特征，便可寻到所有元素
            if func : 
                break
    
    ## 根据 查找到的 索引值 查找所有元素
    def __find(self, index):
        """
        根据基字索引 寻找其他元素
        :param index: 基字 的索引值 
        """
        # 若 基字前只有一个字 
        if index == 1 :
            if self.__str[index] in self.__sub_word :               # 如果基字 为下加形态，判断前项字符为 上加字
                    self.__result[1] = self.__str[index-1]
            else:                                                   # 如果基字 为正常形态，判断前项字符为 前加字
                    self.__result[0] = self.__str[index-1]
                    
        # 若 基字前有两个字 则必为 前加字 + 上加字
        if index == 2 :
            self.__result[1] = self.__str[index-1]
            self.__result[0] = self.__str[index-2]                 
        # 若基字后 有一个字符
        if len(self.__str)-index == 2 :                             
            if self.__str[index+1] in (self.__sub_word  + self.__below): # 可能是 下加字--不在正字规范里判断，需记录错误字符、后续的同理
                self.__result[3] = self.__str[index+1]              
            if self.__str[index+1] in self.__symbol :               # 可能是 音标
                self.__result[4] = self.__str[index+1]              
            if self.__str[index+1] in self.__words :                # 可能是 后加字
                self.__result[5] = self.__str[index+1]              
            if self.__str[index+1] == "ཌ" :                         # 可能是 缩写字
                self.__result[5] = "ག"
                self.__result[6] = "ས"  
        # 若基字后 有两个字符 
        if len(self.__str)-index == 3 :                             
            if self.__str[index+1] in (self.__sub_word + self.__below) :  # 如果 第一个是 下加字
                self.__result[3] = self.__str[index+1]              # 
                if self.__str[index+2] in self.__symbol :           # 下加字 + 音标 的可能性
                    self.__result[4] = self.__str[index+2]          # 
                if self.__str[index+2] in self.__words :            # 下加字 + 后加字的可能性
                    self.__result[5] = self.__str[index+2]          # 
                if self.__str[index+2] == "ཌ" :                     # 可能是 缩写字
                    self.__result[5] = "ག"
                    self.__result[6] = "ས"  
            if self.__str[index+1] in self.__symbol :               # 如果 第一个是 音标 
                if self.__str[index+2] == "ཌ" :                     # 可能是 缩写字
                    self.__result[5] = "ག"
                    self.__result[6] = "ས"  
                else:                                               # 音标 + 后加字
                    self.__result[4] = self.__str[index+1]          # 保存 音标
                    self.__result[5] = self.__str[index+2]          # 保存 后加字
            if self.__str[index+1] in self.__words :                # 如果 第一个是 后加字 第二个是重后加字
                self.__result[5] = self.__str[index+1]              # 保存 后加字
                self.__result[6] = self.__str[index+2]              # 保存 重后加字
        # 若基字后 有三个字符 
        if len(self.__str)-index == 4 :                             
            if self.__str[index+1] in (self.__sub_word  + self.__below):  # 第一个是下加字的情景
                self.__result[3] = self.__str[index+1]
                if self.__str[index+2] in self.__symbol:            
                    if self.__str[index+3] == "ཌ" :                 # 可能是 缩写字
                        self.__result[4] = self.__str[index+2]      # 保存 音标
                        self.__result[5] = "ག"                          
                        self.__result[6] = "ས"     
                    else:                                           # 下加字 + 音标 + 后加字
                        self.__result[4] = self.__str[index+2]      # 保存 音标
                        self.__result[5] = self.__str[index+3]      # 保存 后加字
                if self.__str[index+2] in self.__words:             # 下加字 + 后加字 + 重后加字
                    self.__result[5] = self.__str[index+2]          # 保存 后加字
                    self.__result[6] = self.__str[index+3]          # 保存 重后加字
            if self.__str[index+1] in self.__symbol:                # 音标 + 后加字 + 重后加字的情景
                self.__result[4] = self.__str[index+1]              # 保存 音标
                self.__result[5] = self.__str[index+2]              # 保存 后加字
                self.__result[6] = self.__str[index+3]              # 保存 重后加字
        # 若基字后 有四个字符  
        if len(self.__str)-index == 5 :                             # 必然是 下加字 + 音标 + 后加字 + 重后加字
            self.__result[3] = self.__str[index+1]
            self.__result[4] = self.__str[index+2]
            self.__result[5] = self.__str[index+3]
            self.__result[6] = self.__str[index+4]
    
    ## 查找 基字及其索引值的 策略
    def __if_single(self):
        """
        单字成词 མིང་གཞི་རྐྱང་པ་
        :return true: 如果程序执行返回 true   
        """ 
        if len(self.__str) == 1 :    
            self.__result[2] == self.__str
            
            return True
        
    def __if_top(self):
        """
        མགོ་ཅན་ཡོད་པ་  当存在上加字  寻找基字及其索引值
        :return true: 如果程序执行返回 true
        """
        if set(self.__str) & set(self.__sub_word) : 
            sub_word = [i for i in self.__str  if i in self.__sub_word] # 找到 下加形态的基字
            # 考虑到特殊字的基字不是 下加形态 eg: གྷ
            index_sub = self.__str.index(sub_word[0])                    # 下加形态字符的 索引值
            if sub_word[0] in ["ྙ", "ྥ", "ྷ"]  :
                if  (sub_word[0] == "ྙ") & ((self.__str[index_sub-1]) == "ཛ") :
                    index = index_sub 
                    self.__result[2] = self.__str[index]             # 保存 基字
                    self.__find(index)                               # 寻找其他元素 
                if (sub_word[0] == "ྥ") & ((self.__str[index_sub-1]) == "ཧ") :
                    index = index_sub 
                    self.__result[2] = self.__str[index]             # 保存 基字
                    self.__find(index)                               # 寻找其他元素 
                if (sub_word[0] == "ྷ") & ((self.__str[index_sub-1]) in ["ག", "ད", "ཛ", "བ", "ཧ"]) :
                    index = index_sub -1
                    self.__result[2] = self.__str[index]             # 保存 基字
                    self.__find(index)                               # 寻找其他元素   
                else:
                    index = index_sub
                    self.__result[2] = self.__new_dict[self.__str[index]]       # 保存 基字
                    self.__find(index)                               # 寻找其他元素  
            else:
                index = index_sub                                           # 找到 基字 在字符串中的索引值
                self.__result[2] = self.__new_dict[self.__str[index]]       # 保存 基字
                self.__find(index)                               # 寻找其他元素  
           
            return True
            
    def __if_below(self):
        """
        འདོགས་ཅན་ཡོད་པ་  当存在下加字  寻找基字及其索引值
        :return true: 如果程序执行返回 true
        """
        if set(self.__str) & set(self.__below) : 
            below_word = [i for i in self.__str if i in self.__below]   # 找到 下加字
            index = self.__str.index(below_word[0]) - 1                 # 找到 基字 在字符串中的索引值
            if self.__str[index] in self.__sub_word :                   # 根据基字形态 保存 基字 
                self.__result[2] = self.__new_dict[self.__str[index]]
            else :
                self.__result[2] = self.__str[index]    
            self.__find(index)                               # 寻找其他元素
            
            return True

    def __if_symbol(self):
        "དབྱངས་ཡོད་པ་  当存在音标 寻找基字及其索引值"
        if set(self.__str) & set(self.__symbol) : 
            symbol = [i for i in self.__str if i in self.__symbol]          # 找到 音标
            symbol_index = self.__str.index(symbol[0])                      # 找到 音标的索引值
            if self.__str[symbol_index-1] in self.__below :                 # 当音标前是 下加字时 
                index =  symbol_index-2                                     # 找到 基字 的索引值
                if self.__str[index] in self.__sub_word :                   # 根据基字形态 保存 基字 
                    self.__result[2] = self.__new_dict[self.__str[index]]
                else :
                    self.__result[2] = self.__str[index]    
                self.__find(index) 
            else :
                index = symbol_index -1                                     # 找到 基字 的索引值
                if self.__str[index] in self.__sub_word :                   # 根据基字形态 保存 基字 
                    self.__result[2] = self.__new_dict[self.__str[index]]       
                else :
                    self.__result[2] = self.__str[index]    
                self.__find(index)                               # 寻找其他元素
        return True
    
    def __if_pure_word(self):
        "当字符串里没有上加字、下加字、和音标时 寻找基字及 其他元素"
        state = []                                                          # 保持 字符串里各元素存在的 信息 True为存在 Flase为不存在
        state.append(bool(set(self.__str) & set(self.__symbol)))            # 查看是否存在 音标
        state.append(bool(set(self.__str) & set(self.__sub_word)))          # 查看是否存在 上加字
        state.append(bool(set(self.__str) & set(self.__below)))             # 查看是否存在 下加字
        if True not in state :                                  
            if len(self.__str) == 4 :                                       # 前加字 + 基字 + 后加字 + 重后加字的情景
               self.__result[0] = self.__str[0]  
               self.__result[2] = self.__str[1]
               self.__result[5] = self.__str[2]
               self.__result[6] = self.__str[3]
            if len(self.__str) == 3 :
                if self.__str[-1] == "ཌ" :                                  # 可能是 缩写字
                    self.__result[0] = self.__str[0]  
                    self.__result[2] = self.__str[1]
                    self.__result[5] = "ག"
                    self.__result[6] = "ས"
                if self.__str[-1] not in self.__repeat_behind and self.__str[-1] != "ཌ" :  # 如果最后一个字不是 重后加字， 则：前加字 + 基字 + 后加字
                    self.__result[0] = self.__str[0]
                    self.__result[2] = self.__str[1]
                    self.__result[5] = self.__str[2]
                if self.__str[-1]  in self.__repeat_behind :
                    if self.__str[-1] == "ས":
                        if self.__str[1] in ["ག", "ད", "བ", "མ"] :  # 当只有三个字根时：ས 的前项是这些字时，ས 作为 重后加字 存在
                            self.__result[2] = self.__str[0]
                            self.__result[5] = self.__str[1]
                            self.__result[6] = self.__str[2]
                        else :
                            self.__result[0] = self.__str[0]
                            self.__result[2] = self.__str[1]
                            self.__result[5] = self.__str[2]
                    
                    else :                                          # 重后加字 ད 已被弃用，一般三个字尾出现时，当做后加字
                        self.__result[0] = self.__str[0]
                        self.__result[2] = self.__str[1]
                        self.__result[5] = self.__str[2]    
            if len(self.__str) == 2 :                               # 当只有两个字  一般都是  基字 + 后加字 这样处理
                if self.__str[-1] == "ཌ" :
                    self.__result[2] = self.__str[0]
                    self.__result[5] = "ག"
                    self.__result[6] = "ས"
                else :  
                    if self.__str[0] in self.__front :
                        if self.__str[1] in self.__behind :         # 当第一个字符在 前加字集 且 第二个在后加字集时 eg བས་
                            self.__result[2] = self.__str[0]
                            self.__result[5] = self.__str[1]
                        else:                                       # 当第一个字符在 前加字集 且 第二个 不 在后加字集时 eg བཀ་
                            self.__result[0] = self.__str[0]
                            self.__result[2] = self.__str[1]
                    else :                                          # 当第一个字符不在 前加字集时  eg  ཀས
                        
                        self.__result[2] = self.__str[0]
                        self.__result[5] = self.__str[1]
        return True       
    
    ## 审查 __result 各元素合法性函数，供外调函数使用
    def __check_front(self):
        """
        审查 前加字
        :return str/None:  返回错误信息, 字符合法返回None
        """
        if self.__result[0] is not None :  
            if self.__result[0] not in self.__front :
                return "不存在前加字(སྔོན་འཇུག་) {}".format(self.__result[0])
    
    def __check_top(self):
        """
        审查 上加字
        :return str/None:  返回错误信息, 字符合法返回None
        """
        if self.__result[1] is not None :
            if self.__result[1]  in self.__top:
                
                if self.__result[1] + self.__result[2] not in self.__special_words :  # 若不是特殊字、则分别查看正字规范表
                    if self.__result[1] == "ར" :
                        if self.__result[2] not in self.__top_ra :
                            return ("上加字ra(མགོ་ཅན་ ར)的正字规范里不存在字符： {0} ！！".
                                    format(self.__result[1]+self.__sub_word_dict[self.__result[2]]))    # 需要把基字转换成 下加形态打印
                    if self.__result[1] == "ལ" :
                        if self.__result[2] not in self.__top_la :
                            return ("上加字la(མགོ་ཅན་ ལ)的正字规范里不存在字符： {0} ！！".
                                    format(self.__result[1]+self.__sub_word_dict[self.__result[2]]))    # 需要把基字转换成 下加形态打印
                    if self.__result[1] == "ས" : 
                        if self.__result[2] not in self.__top_sa :
                            return ("上加字sa的正字规(མགོ་ཅན་ ས)范里不存在字符： {0} ！！".
                                    format(self.__result[1]+self.__sub_word_dict[self.__result[2]]))    # 需要把基字转换成 下加形态打印
                else :      # 既不是特殊字 便是 也不是ra la sa 则为非法字  
                    return ("不存在上加字(མགོ་ཅན་)： {}".format(self.__result[1]))
            else :
                return ("不存在上加字(མགོ་ཅན་)： {}".format(self.__result[1]))
        
    def __ckeck_under(self):
        """
        审查 下加字
        :return str/None:  返回错误信息, 字符合法返回None
        """
        if self.__result[3] is not None :
            if self.__result[3] not in self.__below :                                             # 当不在 正字规范里时                              
                if (self.__result[2] + self.__result[3]) not in self.__special_words :            # 若不是特殊字、则分别查看正字规范表
                    return ("不存在下加字(འདོགས་ཅན་)： {}".format(self.__new_dict[self.__result[3]]))  # 由于是非法的下加字 需要转换一下形态
            else:
                if self.__result[3] == "ྱ" :
                    if self.__result[2] not in self.__under_ya :
                        return ("下加字 (འདོགས་ཅན་ ཡ་རྟགས་) ྱ 的规范里不存在字符: {}".format(self.__result[2]+self.__result[3]))
                if self.__result[3] == "ྲ" :
                    if self.__result[2] not in self.__under_ra :
                        return ("下加字 (འདོགས་ཅན་ ར་རྟགས་) ྱ 的规范里不存在字符: {}".format(self.__result[2]+self.__result[3]))
                if self.__result[3] == "ླ" :
                    if self.__result[2] not in self.__under_la :
                        return ("下加字 (འདོགས་ཅན་ ལ་རྟགས་) ྱ 的规范里不存在字符: {}".format(self.__result[2]+self.__result[3]))
                if self.__result[3] == "ྭ" :
                    if self.__result[2] not in self.__under_wa :
                        return ("下加字 (འདོགས་ཅན་ ཝ་རྟགས་) ྱ 的规范里不存在字符: {}".format(self.__result[2]+self.__result[3]))
    
    def __check_behind(self):
        """
        审查 后加字
        :return str/None:  返回错误信息, 字符合法返回None
        """
        if self.__result[5] is not None :
            if self.__result[5] not in self.__behind :
                return ("不存在后加字(རྗེས་འཇུག་): {}".format(self.__result[5]))
    
    def __check_repeat_behind(self):
        """
        审查 重后加字
        :return str/None:  返回错误信息, 字符合法返回None
        """
        if self.__result[6] is not None :
            if self.__result[6] not in ["ད", 'ས'] :
                return ("不存在重后加字(ཡང་འཇུག་): {}".format(self.__result[6]))
            else :
                if self.__result[6] == "ད":
                   if self.__result[5] not in ["ན", "ར", "ལ"] :
                       return ("重后加字(ཡང་འཇུག་) ད 使用不当，属于非法字符")
                if self.__result[6] == "ས":
                    if self.__result[5] not in ["ག", "ང", "བ", "མ"] :
                       return ("重后加字(ཡང་འཇུག་) ས 使用不当，属于非法字符")

    #############################################################################
    # 外调函数
    def front(self):
        """
        :return str: སྔོན་འཇུག་ 获取前加字
        """
        self.__analysis()
        return self.__result[0]
    
    def top(self):
        """
        :return str: མགོ་ཅན་ 获取上加字
        """
        self.__analysis()
        return self.__result[1]

    def basic(self): 
        """
        :return str: མིང་གཞི་ 获取基字
        """
        self.__analysis()
        return self.__result[2]

    def below(self): 
        """
        :return str: འདོགས་ཅན་ 下加字
        """
        self.__analysis()
        return self.__result[3]
    
    def symbol(self):
        """
        :return str: དབྱངས་  音标
        """
        self.__analysis()
        return self.__result[4]
    
    def behind(self): 
        """
        :return str: རྗེས་འཇུག་ 后加字
        """
        self.__analysis()
        return self.__result[5]

    def repeat_behind(self):
        """
        :return str: ཡང་འཇུག་ 重后加字
        """
        self.__analysis()
        return self.__result[6]
    
    def get_all(self):
        """
        :return list: 元素映射表：[前加字, 上加字, 基字, 下加字, 后加字, 重后加字]
        """
        self.__analysis()
        return self.__result

    def is_right(self):
        """
        根据正字规范，判断词的合法性
        :return str: 返回错误信息 
        """
        # 1.分析字符串 并填充 信息到数组
        self.__analysis()                
        # 2.将审查信息集存进一个 list 中
        results = [self.__check_front(), self.__check_top(), self.__ckeck_under(), self.__check_behind(), self.__check_repeat_behind()]
        # 3.观察列表
        state = True                                # 用于存储 list 元素信息
        for result in results :
            if result != None :
                state = False                       # 只要其中一个不是None将状态值切换并，跳出循环
                break
        # 4. 处理列表
        data = ["该字符 "]                          # 存储 反馈信息
        if state == False:                          # 若 state 为 False，则字符存在 非法元素
            for result in results:
                if result != None :                 # 过滤 None
                    data.append(result)             
        else:
            data.append("合法！")
        
        # 5.制作函数返回值 - 此处采用策略：转成str、并去掉list和str的边框
        return str(data)[1:-1].replace("'", '').replace(",","---")
      
#############################################################################
# test
# s = Word("བརྒགས")
# s = Word("བརྒྲཌ")
# s = Word("བརྒྱོཌ")

# s = Word("ལགས་")
# s = Word("བསྒྲིགས")

# s = Word("ལྣཏས")

# print("basic         ", s.basic())
# print("behind:        ", s.behind())
# print("below:         ", s.below())
# print("front:         ", s.front())
# print("repeat_behind: ", s.repeat_behind())
# print("top:           ", s.top())
# print("get_all:       ", s.get_all())
# print("symbol:        ", s.symbol())

# print(s.is_right())